#!/usr/bin/perl
#
#
#
###############################################################################
# Script:	cruft-control.pl
# Version:	0.23-dev
# Autor:	Daniel Pielmeier
###############################################################################

use strict;
use diagnostics;
use warnings;
use File::Find;
#use Text::Diff;
use Getopt::Long qw (:config posix_default gnu_compat pass_through
		no_ignore_case_always);


# Directories

my $destdir = ();
my $confdir = ();
my $searchdir = ();
my $vdb_path = ();


# Files

my $cruft = "cruft.txt";
my $oldcruft = "oldcruft.txt";
my $diffcruft = "diffcruft.txt";
my $negcruft = "negcruft.txt";
my $systemfile = "systemfile.txt";
my $contentlist = "contentlist.txt";
my $packagefile = "packagefile.txt";
my $ignorelist = "ignorelist.conf";


# Variables

my $uid = `id -u`; chomp $uid;
my $debug = ();
my $ignoreitem = ();
my @systemfile = my @ignorelist = ();
my @contentlist = ();
my @packagefile = ();
my $prev = ();
my @cruft = ();
my %packagefilecheck = ();
my @negcruft =();
my %systemfilecheck = ();
my $help = "
DESCRIPTION:
This script tries to find files not managed by the package manager. The
resulting list of files should normally not called cruft as proper detection
of cruft is almost impossible. So this script does not offer functions to
delete any files reported.

In general it searches all files in the filesystem. It uses an
$ignorelist where files and directories can be specified which will be
ignored from the search to avoid obvious false positives.
Then it gets the files which are listed in CONTENTS from vdb_path where the
files installed by the package manager are stored. This two resulting lists
are compared with each other to get the files which are not under the control
of the package manager. The resulting list is stored in the $cruft file.

Additionally if available it compares the results with an older cruft file
called $oldcruft and creates $diffcruft out of it to find out changes
which occur when merging and unmerging programs. This is useful when you are
not sure about the files listed in an initial search so one can just track
the changes.

Carefully examine $cruft and/or $diffcruft. Deleting wrong files which
are essential for a working system may break it. You are responsible for your
system. This script will just make it easier to to detect files which are
probably not needed on the system.

Optionally it writes debug information in form of $systemfile, $contentlist,
$packagefile and $negcruft.

$systemfile lists all files found in the filesystem excluding the files
and directories from $ignorelist.
$contentlist lists all paths to the CONTENT files only useful to see which
packages are installed.
$packagefile lists all files under the control of the package manager.
$negcruft this are the files which are in $packagefile but not in
$systemfile because they are not present in the filessystem or ignored by
the ignorefile.

CONFIGURATION:
For $0 you can configure the path to configuration files which
for now is only $ignorelist. and the path to generated output ($cruft,
$oldcruft, $diffcruft) and debug files ($systemfile, $packagefile,
$contentlist, $negcruft).

By default the generated output and debug files go to the current working
directory and configuration files are read from /etc/cruft-control
This paths can be overridden by the environment variables CRUFT_DESTDIR for
generated output and debug files and CRUFT_CONFDIR for configuration files
or on the commandline using the relevant switches.

Order of preference = default settings < environment variable < command line

DEPENDENCIES:
dev-lang/perl
modules: File::Find, Getopt::Long
sys-apps/portage: portageq for getting the vdb_path
sys-apps/diffutils: diff for comparing old and new results
alternative to sys-apps/diffutils (Text::Diff (Algorithm::Diff)) currently
deactivated

TODO:
- Implement POD::USAGE.
- Decide wether to include searchpath and pm database in the available
  options or not.
- Decide wether to use sys-apps/diffutils or Text::Diff in combination with
  Algorithm::Diff
- Implement perl header workaround to exclude generated perl header files
  created from the corresponding C Header files in /usr/include similar to the
  py{c,o} workaround
- Check if the symlinks from the $cruft file which are currently ignored are
  valid, if not the symbolic link should be considered as cruft
";


# Configurable paths

if (defined $ENV{"CRUFT_DESTDIR"}) {
	$destdir = $ENV{"CRUFT_DESTDIR"};
} else {
	$destdir = ("./");
}

if (defined $ENV{"CRUFT_CONFDIR"}) {
	$confdir = $ENV{"CRUFT_CONFDIR"};
} else {
	$confdir = ("/etc/cruft-control");
}

# Path to the portage database which holds information of all installed ebuilds.
$vdb_path = qx (portageq vdb_path);
chomp ($vdb_path);


# Hardcoded searchdir

# Path to the searchdir. Normally this is the root-dir if you want to search the
# whole system.
$searchdir = ("/");


# Checking for permissions and correct usage so the program can access all
# files and directories.

if ($uid ne "0") {
	print "\e[1;31m[ERROR]\tYou need to be root to use $0 for accessing all\n";
	print "[ERROR]\tdirectories and files!\e[m\n\n";
	&usage();
}

GetOptions ('destdir|d=s' => \$destdir, 'confdir|c=s' => \$confdir,
			#'searchdir|s=s' => \$searchdir, 'vdb_path|v=s' => \$vdb_path,
			'debug|b' => \$debug, 'help|h|?' => sub {usage()});

&usage() if ($#ARGV > -1);

foreach ("$destdir", "$confdir") {
	unless (-d $_) {
		print "\e[1;31m[ERROR]\tThe directory $_ does not exist\e[m\n\n";
	&usage();
	}
}


# Creating a list of all files in the filesystem excluding files and
# directories which belong to an ignorelist.

print "[INFO]\tSearching files on the filesystem ...\n";

if (-f "$confdir/$ignorelist" && -s _ && -T _) {
	open (IGNORELIST,"<$confdir/$ignorelist") || die ("$ignorelist: $!");

	print "[INFO]\tExcluding files from $confdir/$ignorelist\n";

	foreach (<IGNORELIST>) {
		unless ( $_ eq "\n" || $_ =~ "^#" ) {
			push (@ignorelist, $_);
		}
	}
	chomp (@ignorelist);
	
	find(\&sysignore, $searchdir);
	sub sysignore() {
		my $count = 0;

		foreach $ignoreitem (@ignorelist) {
			if ("$File::Find::name" eq "$ignoreitem") {
				splice(@ignorelist,$count,1);
				($File::Find::prune = 1);
				return;
			}
			$count++;
		}
		push (@systemfile, "$File::Find::name\n");
	}
} else {
	print "\e[1;33m[WARN]\tThe file $ignorelist does not exist in\n[WARN]\t";
	print "$confdir\n[WARN]\tor has no content. This may result in a big and ";
	print "probably\n[WARN]\tuseless cruft list. ;-)\e[m\n";

	find(\&sys, $searchdir);
	sub sys() {
		push (@systemfile, "$File::Find::name\n");
	}
}

@systemfile = sort (@systemfile); 


# Trying to workaround the byte compiled python modules which are created by
# python_mod_optimize/cleanup in pkg_postinst()/pkg_postrm() so the {pyc,pyo}
# files are not recorded to CONTENTS
# Better would be checking if a pyc or pyo file from @systemfile has a
# corresponding py file in @packagefile, but simply checking in @systemfile
# only should be sufficient.

print "[INFO]\tWorkaround for byte compiled python modules ...\n";

my $syscount;
my $removed;
my $py;

foreach (@systemfile) {
	if ( $_ =~ /py(c$|o$)/) {
		$py = substr($_, 0, -2) . "\n";
		if ($systemfile[$syscount-1] eq $py) {
			$removed = splice (@systemfile,$syscount,1,$py);
		}
	}
	$syscount++;
	#print "$syscount\n";
}

$prev = "not equal to $systemfile[0]";
@systemfile = grep($_ ne $prev && ($prev = $_, 1), @systemfile);


# Searches the package database for the CONTENTS-files of all installed
# packages. This is the base for getting the list of managed files.

print "[INFO]\tSearching installed packages ...\n";

find(\&content, $vdb_path);
sub content() {
	push (@contentlist, "$File::Find::name\n") if (/CONTENTS/);
}

@contentlist = sort (@contentlist);


# Creating a list of all files owned by all installed packages.

print "[INFO]\tAssigning files to installed packages ...\n";

foreach (@contentlist) { 
	open (CONTENT, "$_") || die("$_: $!");
	foreach (<CONTENT>) {
		if ((/^sym/) || (/^dir/)) {
			push (@packagefile, "$2\n") if (/(\w+\s)(\S+)/);
		}
		if (/^obj/) {
			push (@packagefile, "$2\n") if (/(\w+\s)(.+)(\s\w+\s\d+$)/);
		}
	}
}

@packagefile = sort (@packagefile);
$prev = "not equal to $packagefile[0]";
@packagefile = grep($_ ne $prev && ($prev = $_, 1), @packagefile);


# Compares the list of the files in the system with the files managed by
# portage to get unmanaged files. It excludes symlinks as the are mostly
# created in pkg_posintall, or at rutime by eselect or other tools.

print "[INFO]\tComparing to get unmanaged files ...\n";

open (CRUFT,">$destdir/$cruft") || die ("$cruft: $!");

@packagefilecheck{@packagefile} = ();

foreach $_ (@systemfile) {
	unless ( exists $packagefilecheck{$_} ) {
		chomp($_);
		push (@cruft, "$_\n") unless -l $_;
	}
}

@cruft = sort (@cruft);
print CRUFT @cruft;

print "[INFO]\tList containing cruft written to $destdir/$cruft\n";


# Compares the list of found cruft files with a list of files found in an
# older search.

if (-f "$destdir/$oldcruft" && -s _ && -T _) {
	print "[INFO]\tComparing cruft with old results ...\n";

	#open (OLDCRUFT,"<$destdir/$oldcruft") || die ("$oldcruft: $!");
	#my @oldcruft = <OLDCRUFT>;
	#print "[INFO]\tUsing previous results from $destdir/$oldcruft\n";
	#open (DIFFCRUFT,">$destdir/$diffcruft") || die ("$diffcruft: $!");
	#diff \@oldcruft, \@cruft, {OUTPUT => \*DIFFCRUFT};
	system "diff -u $destdir/oldcruft.txt $destdir/cruft.txt >$destdir/diffcruft.txt";

	print "[INFO]\tDifferences to previous run written to ";
	print "$destdir/$diffcruft\n";
} else {
	print "[INFO]\tThe file $oldcruft does not exist in\n";
	print "[INFO]\t$destdir\n[INFO]\tor has no contents.\n";
	print "[INFO]\tNote: For later comparisons you can move $cruft to ";
	print "$oldcruft\n[INFO]\tto catch the differences in the file ";
	print "$diffcruft.\n[INFO]\tSkipping comparison with old results ...\n\n";
}


# Write debug files.

&debug() if defined $debug;


# Subroutines

# Print usage message.

sub usage() {
	print "USAGE: $0 [options]\n";
	print "  -d|--destdir=DIR\tdirectory for analysis and debug output\n";
	print "\t\t\t[$destdir]\n";
	print "  -c|--confdir=DIR\tdirectory containing $ignorelist\n";
	print "\t\t\t[$confdir]\n";
	#print "  -s|--searchdir=DIR\tbase directory for searching\n";
	#print "\t\t\t[$searchdir]\n";
	#print "  -v|--vdb_path=DIR\tpath to the portage database\n";
	#print "\t\t\t[$vdb_path]\n";
	print "  -b|--debug\t\tgenerate debug output\n";
	print "  -?|-h|--help\t\tprint this lot out\n";
	print $help;
	exit;
}


# Subroutine to generate debug output.

sub debug() {

	open (SYS,">$destdir/$systemfile") || die ("$systemfile: $!");
	print SYS @systemfile;
	print "\e[1;35m[DEBUG]\tAll files from the system excluding the ignored ";
	print "files written to $destdir/$systemfile\e[m\n";

	open (CONTENT,">$destdir/$contentlist") || die ("$contentlist: $!");
	print CONTENT @contentlist;
	print "\e[1;35m[DEBUG]\tPath to files containing package contents ";
	print "written to $destdir/$contentlist\e[m\n";

	open (PACKAGE,">$destdir/$packagefile") || die ("$packagefile: $!");
	print PACKAGE @packagefile;
	print "\e[1;35m[DEBUG]\tAll files managed by the package manager ";
	print "written to $destdir/$packagefile\e[m\n";

# For reference list the files which are managed by portage but not in the
# system maybe because they are deleted or ignored by the ignorefile.

	print "\e[1;35m[DEBUG]\tComparing to get negcruft files ...\n";

	@systemfilecheck{@systemfile} = ();

	foreach $_ (@packagefile) {
		push (@negcruft, $_) unless exists $systemfilecheck{$_};
	}

	@negcruft = sort (@negcruft);

	open (NEGCRUFT,">$destdir/$negcruft") || die ("$negcruft: $!");
	print NEGCRUFT @negcruft;
	print "\e[1;35m[DEBUG]\tNegative list of cruft written to ";
	print "$destdir/$negcruft\e[m\n";
}
